# Standard library
import re
from collections import defaultdict

# Third party
from saplings.abstract import Tool

# Local
try:
    from suss.reranker import rerank_chunks
    from suss.index import Index, File, Chunk
    from suss.constants import CODE_SEARCH_LIMIT
except ImportError:
    from reranker import rerank_chunks
    from index import Index, File, Chunk
    from constants import CODE_SEARCH_LIMIT


#########
# HELPERS
#########


def query_to_regex(query: str) -> str:
    return "|".join(map(re.escape, query.split()))


async def filter_chunks(query: str, chunks: list[Chunk]) -> list[Chunk]:
    return chunks  # TODO: LLM-based filtering


######
# MAIN
######


class GrepCodeTool(Tool):
    def __init__(
        self, index: Index, target_file: File, update_progress: callable, **kwargs
    ):
        # Base attributes
        self.name = "search_code"
        self.description = "Searches the contents of files in the codebase using regular expressions. Returns code snippets containing exact matches."
        self.parameters = {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Concise, one-sentence description of your intent behind the search. E.g. 'Find the definition of handle_auth', 'Track lifecycle of connection_pool', 'Understand how the parser is used'.",
                },
                "query": {
                    "type": "string",
                    "description": "A search query, passed into Python's re.match(). Should match symbols in the codebase.",
                },
            },
            "required": ["intent", "query"],
            "additionalProperties": False,
        }
        self.is_terminal = False

        # Additional attributes
        self.index = index
        self.target_file = target_file
        self.update_progress = update_progress

    def format_output(self, chunks: list[Chunk]) -> str:
        grouped_chunks = defaultdict(list)
        for chunk in chunks:
            grouped_chunks[chunk.file].append(chunk)

        formatted_chunks = []
        for file, chunks in grouped_chunks.items():
            line_nums = set()
            for chunk in chunks:
                line_nums |= set(chunk.line_nums)
            line_nums = list(line_nums)
            line_nums.sort()

            chunk = Chunk(line_nums, file)
            formatted_chunk = f"<file_path>{file.rel_path}</file_path>\n<file_content>\n{chunk.to_string()}\n</file_content>"
            formatted_chunks.append(formatted_chunk)

        formatted_chunks = "\n\n".join(formatted_chunks)
        return formatted_chunks

    async def run(self, intent: str, query: str, **kwargs) -> list[Chunk]:
        self.update_progress(intent)
        query_regex = query_to_regex(query)
        results = self.index.search_code(query_regex, exclude=self.target_file)
        results = await rerank_chunks(query, results)
        results = await filter_chunks(query, results)
        return results[:CODE_SEARCH_LIMIT]


# TODO: Implement a semantic fallback if AST-grep fails to retrieve enough chunks.
# 1. Get the code map (c-tags) for the codebase.
# 2. Use the code map and the query to generate a hypothetical code snippet.
# 3. Run an (AST-based) keyword search on the codebase using the snippet.
# 4. Rerank the results.



# Standard library
from dataclasses import dataclass
from collections import defaultdict

# Third party
import json_repair
from saplings.dtos import Message
from saplings.abstract import Tool
from litellm import acompletion, encode
from sortedcollections import OrderedSet

# Local
try:
    from suss.index import Chunk, File
    from suss.constants import (
        MAX_MERGE_DISTANCE,
        MAX_CONTEXT_TOKENS,
        MAX_THINKING_TOKENS,
    )
except ImportError:
    from index import Chunk, File
    from constants import MAX_MERGE_DISTANCE, MAX_CONTEXT_TOKENS, MAX_THINKING_TOKENS


#########
# HELPERS
#########


@dataclass
class Bug:
    start: int
    end: int
    description: str
    confidence: float


SYSTEM_PROMPT = """I want to find all the bugs in a code file, if any exist. A bug is \
code that could cause runtime errors, crashes, or incorrect behavior. It is NOT a style \
issue, suggestion, or anything else.

Your output should resemble an expert code review. It should be a list of JSON objects, \
one for each bug, each containing a code block and a description of the bug contained in \
that code block.

Remember that you are analyzing a code file for bugs. It is entirely possible that the code \
file contains no bugs, in which case you should return an empty list.

--

The code file you're analyzing is part of a codebase. Below is additional context from the \
codebase that may help you understand the code file. Use it to help you identify bugs, if \
necessary.

<context>
{context}
</context>"""


def get_reasoning_model(model: str) -> str:
    model = model.lower()
    if model.startswith("openai/"):
        return "openai/o3-mini"
    elif model.startswith("anthropic/"):
        return "anthropic/claude-3-7-sonnet-20250219"

    return model


def group_chunks_by_file(chunks: list[Chunk]) -> dict[File, list[Chunk]]:
    chunks_by_file = defaultdict(list)
    for chunk in chunks:
        chunks_by_file[chunk.file].append(chunk)

    return chunks_by_file


def get_chunks(trajectory: list[Message]) -> OrderedSet[Chunk]:
    chunks = OrderedSet()
    for message in trajectory:
        if not message.raw_output:
            continue

        if not hasattr(message.raw_output, "__iter__"):
            continue

        for item in message.raw_output:
            if isinstance(item, Chunk):
                chunks.add(item)

    return chunks


def filter_chunks(chunks: list[Chunk], files: list[str]) -> list[Chunk]:
    return [chunk for chunk in chunks if chunk.file.rel_path in files]


def get_contiguous_subchunks(line_nums: list[int], file: File) -> list[Chunk]:
    if not line_nums:
        return []

    groups = []
    curr_group = [line_nums[0]]
    for line_num in line_nums[1:]:
        if line_num == curr_group[-1] + 1:
            curr_group.append(line_num)
        else:
            groups.append(curr_group)
            curr_group = [line_num]

    if curr_group:
        groups.append(curr_group)

    chunks = [Chunk(group, file) for group in groups]
    return chunks


def merge_chunks(chunks: list[Chunk]) -> list[Chunk]:
    chunks_by_file = group_chunks_by_file(chunks)
    merged_chunks = []
    for file, chunks in chunks_by_file.items():
        all_line_nums = {ln for chunk in chunks for ln in chunk.line_nums}
        all_line_nums = list(sorted(all_line_nums))
        merged_chunks += get_contiguous_subchunks(all_line_nums, file)

    return merged_chunks


def truncate_chunks(chunks: list[Chunk], model: str) -> list[Chunk]:
    num_tokens = 0
    truncated_chunks = []
    for chunk in chunks:
        tokens = len(encode(model=model, text=chunk.to_string()))
        if num_tokens + tokens > MAX_CONTEXT_TOKENS:
            break

        num_tokens += tokens
        truncated_chunks.append(chunk)

    return truncated_chunks


def normalize_chunks(chunks: list[Chunk]) -> list[Chunk]:
    chunks_by_file = group_chunks_by_file(chunks)
    norm_chunks = []
    for file, chunks in chunks_by_file.items():
        all_line_nums = {ln for chunk in chunks for ln in chunk.line_nums}
        all_line_nums = list(sorted(all_line_nums))

        norm_line_nums = []
        for index in range(len(all_line_nums)):
            curr_line_num = all_line_nums[index]
            next_line_num = (
                all_line_nums[index + 1] if index + 1 < len(all_line_nums) else None
            )

            norm_line_nums.append(curr_line_num)
            for i in range(1, MAX_MERGE_DISTANCE + 1):
                if not next_line_num:
                    if curr_line_num + i > file.last_lineno:
                        break

                if curr_line_num + i == next_line_num:
                    break

                norm_line_nums.append(curr_line_num + i)

        norm_chunks.append(Chunk(norm_line_nums, file))

    return norm_chunks


def build_context_str(chunks: list[Chunk]) -> str:
    fp_str = "<file_paths>\n"
    chunks_str = "<code_chunks>\n"
    for chunk in chunks:
        fp_str += f"{chunk.file.rel_path}\n"
        chunks_str += f"<{chunk.file.rel_path}>\n{chunk.to_string(line_nums=False)}\n</{chunk.file.rel_path}>\n\n"

    chunks_str = chunks_str.rstrip("\n")
    chunks_str += "\n</code_chunks>"
    fp_str += "</file_paths>"

    return f"{fp_str}\n\n{chunks_str}"


async def find_bugs(context: str, file: File, model: str) -> list[Bug]:
    system_message = {
        "role": "system",
        "content": SYSTEM_PROMPT.format(context=context),
    }
    user_message = {
        "role": "user",
        "content": f"<path>{file.rel_path}</path>\n<code>\n{file.content}\n</code>\n\n--\n\nAnalyze the file above for bugs.",
    }
    response = await acompletion(
        model=get_reasoning_model(model),
        messages=[system_message, user_message],
        # frequency_penalty=0.0,
        # temperature=0.75,
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "bug_report",
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {
                        "bugs": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "start": {
                                        "type": "integer",
                                        "description": "Starting line number (inclusive) for the buggy code block.",
                                    },
                                    "end": {
                                        "type": "integer",
                                        "description": "Ending line number (inclusive) for the buggy code block.",
                                    },
                                    "description": {
                                        "type": "string",
                                        "description": "Concise description of the bug and its impact. No more than a few short sentences.",
                                    },
                                    "confidence": {
                                        "type": "integer",
                                        "description": "Confidence score between 0 and 100. 100 indicates high confidence in the bug's existence and severity (i.e. it definitely exists and is severe). 0 indicates low confidence (i.e. bug doesn't exist or is very low severity).",
                                    },
                                },
                                "required": [
                                    "start",
                                    "end",
                                    "description",
                                    "confidence",
                                ],
                                "additionalProperties": False,
                            },
                            "description": "List of bugs in the code file. This list can be empty if there are no bugs in the file.",
                        },
                    },
                    "required": ["bugs"],
                    "additionalProperties": False,
                },
            },
        },
        thinking={"type": "enabled", "budget_tokens": MAX_THINKING_TOKENS},
        drop_params=True,
    )
    response = response.choices[0].message.content
    response = json_repair.loads(response)

    bugs = []
    for bug in response["bugs"]:
        start = max(1, bug["start"])
        end = min(bug["end"], file.last_lineno)
        end = end - 1 if not file.lines[end - 1].strip() else end
        conf_score = max(0, min(100, bug["confidence"]))
        bugs.append(Bug(start, end, bug["description"], conf_score))

    return bugs


######
# MAIN
######


class FindBugsTool(Tool):
    def __init__(
        self, model: str, target_file: File, update_progress: callable, **kwargs
    ):
        # Base attributes
        self.name = "done"
        self.description = (
            "Call this when you have enough context to answer the user's query."
        )
        self.parameters = {
            "type": "object",
            "properties": {
                "files": {
                    "type": "array",
                    "items": {"type": "string", "enum": []},
                    "description": f"Files that contain the context you need to answer the user's query.",
                },
            },
            "required": ["files"],
            "additionalProperties": False,
        }
        self.is_terminal = True

        # Additional attributes
        self.model = model
        self.target_file = target_file
        self.update_progress = update_progress

    def update_definition(self, trajectory: list[Message] = [], **kwargs):
        files = set()
        for message in trajectory:
            if not message.raw_output:
                continue

            if not hasattr(message.raw_output, "__iter__"):
                continue

            for item in message.raw_output:
                if isinstance(item, Chunk):
                    files.add(item.file.rel_path)
                elif isinstance(item, File):
                    files.add(item.rel_path)

        files = [file for file in files if file != self.target_file.rel_path]
        self.parameters["properties"]["files"]["items"]["enum"] = files

    async def run(self, files: list[str], **kwargs) -> list[Bug]:
        trajectory = kwargs.get("trajectory", [])
        chunks = get_chunks(trajectory)
        chunks = filter_chunks(chunks, files)
        chunks = merge_chunks(chunks)
        chunks = truncate_chunks(chunks, self.model)
        chunks = normalize_chunks(chunks)
        context = build_context_str(chunks)
        bugs = await find_bugs(context, self.target_file, self.model)

        return bugs



BOILERPLATE_REGEXES = [
    # Byte-compiled / optimized / DLL files
    "(^|/)__pycache__/",
    # Distribution / packaging
    "(^|/)build/",
    "(^|/)dist/",
    "(^|/)develop-eggs/",
    "(^|/)downloads/",
    "(^|/)eggs/",
    "(^|/)\.eggs/",
    "(^|/)lib/",
    "(^|/)lib64/",
    "(^|/)parts/",
    "(^|/)sdist/",
    "(^|/)var/",
    "(^|/)wheels/",
    "(^|/)pip-wheel-metadata/",
    "(^|/)share/python-wheels/",
    "(^|/)\.egg-info/",
    # Unit test / coverage reports
    "(^|/)htmlcov/",
    "(^|/)\.tox/",
    "(^|/)\.nox/",
    "(^|/)\.hypothesis/",
    "(^|/)\.pytest_cache/",
    # PyBuilder
    "(^|/)target/",
    # IPython
    "(^|/)profile_default/",
    "(^|/)ipython_config\.py$",
    # PEP 582; used by e.g. github.com/David-OConnor/pyflow
    "(^|/)__pypackages__/",
    # MyPy
    "(^|/)\.mypy_cache/",
    # Pyre type checker
    "(^|/)\.pyre/",
    # Caches
    "(^|/)cache/",
    # Dependencies
    "^[Dd]ependencies/",
    # Distributions
    # C deps
    "^deps/",
    "(^|/)configure$",
    # .NET Core Install Scripts
    "(^|/)dotnet-install\.(ps1|sh)$",
    # Linters
    "(^|/)cpplint\.py",
    # Node dependencies
    "(^|/)node_modules/",
    # Next
    "(^|/)\.next/",
    "(^|/)out/",
    "(^|/)dist/",
    "(^|/)cache/",
    # Yarn 2
    "(^|/)\.yarn/releases/",
    "(^|/)\.yarn/plugins/",
    "(^|/)\.yarn/sdks/",
    "(^|/)\.yarn/versions/",
    "(^|/)\.yarn/unplugged/",
    # esy.sh dependencies
    "(^|/)_esy$",
    # Bower Components
    "(^|/)bower_components/",
    # Erlang bundles
    "^rebar$",
    # Go dependencies
    "(^|/)Godeps/_workspace/",
    # Go fixtures
    "(^|/)testdata/",
    # Bootstrap css and js
    "(^|/)bootstrap([^/.]*)(?=\.).*\.(js|css|less|scss|styl)$",
    "(^|/)custom\.bootstrap([^\s]*)(js|css|less|scss|styl)$",
    # Select2
    "(^|/)select2/.*\.(css|scss|js)$",
    # Bulma css
    "(^|/)bulma\.(css|sass|scss)$",
    # Vendored dependencies
    "(3rd|[Tt]hird)[-_]?[Pp]arty/",
    "(^|/)vendors?/",
    "(^|/)[Ee]xtern(als?)?/",
    "(^|/)[Vv]+endor/",
    # Debian packaging
    "^debian/",
    # Haxelib projects often contain a neko bytecode file named run.n
    "(^|/)run\.n$",
    # Bootstrap Datepicker
    "(^|/)bootstrap-datepicker/",
    ## Commonly Bundled JavaScript frameworks ##
    # jQuery
    "(^|/)jquery([^.]*)\.js$",
    "(^|/)jquery\-\d\.\d+(\.\d+)?\.js$",
    # jQuery UI
    "(^|/)jquery\-ui(\-\d\.\d+(\.\d+)?)?(\.\w+)?\.(js|css)$",
    "(^|/)jquery\.(ui|effects)\.([^.]*)\.(js|css)$",
    # jQuery Gantt
    "(^|/)jquery\.fn\.gantt\.js",
    # jQuery fancyBox
    "(^|/)jquery\.fancybox\.(js|css)",
    # Fuel UX
    "(^|/)fuelux\.js",
    # jQuery File Upload
    "(^|/)jquery\.fileupload(-\w+)?\.js$",
    # jQuery dataTables
    "(^|/)jquery\.dataTables\.js",
    # bootboxjs
    "(^|/)bootbox\.js",
    # pdf-worker
    "(^|/)pdf\.worker\.js",
    # Slick
    "(^|/)slick\.\w+.js$",
    # Leaflet plugins
    "(^|/)Leaflet\.Coordinates-\d+\.\d+\.\d+\.src\.js$",
    "(^|/)leaflet\.draw-src\.js",
    "(^|/)leaflet\.draw\.css",
    "(^|/)Control\.FullScreen\.css",
    "(^|/)Control\.FullScreen\.js",
    "(^|/)leaflet\.spin\.js",
    "(^|/)wicket-leaflet\.js",
    # Sublime Text workspace files
    "(^|/)\.sublime-project",
    "(^|/)\.sublime-workspace",
    # VS Code workspace files
    "(^|/)\.vscode/",
    # Prototype
    "(^|/)prototype(.*)\.js$",
    "(^|/)effects\.js$",
    "(^|/)controls\.js$",
    "(^|/)dragdrop\.js$",
    # Typescript definition files
    "(.*?)\.d\.ts$",
    # MooTools
    "(^|/)mootools([^.]*)\d+\.\d+.\d+([^.]*)\.js$",
    # Dojo
    "(^|/)dojo\.js$",
    # MochiKit
    "(^|/)MochiKit\.js$",
    # YUI
    "(^|/)yahoo-([^.]*)\.js$",
    "(^|/)yui([^.]*)\.js$",
    # WYS editors
    "(^|/)ckeditor\.js$",
    "(^|/)tiny_mce([^.]*)\.js$",
    "(^|/)tiny_mce/(langs|plugins|themes|utils)",
    # Ace Editor
    "(^|/)ace-builds/",
    # Fontello CSS files
    "(^|/)fontello(.*?)\.css$",
    # MathJax
    "(^|/)MathJax/",
    # Chart.js
    "(^|/)Chart\.js$",
    # CodeMirror
    "(^|/)[Cc]ode[Mm]irror/(\d+\.\d+/)?(lib|mode|theme|addon|keymap|demo)",
    # SyntaxHighlighter - http://alexgorbatchev.com/
    "(^|/)shBrush([^.]*)\.js$",
    "(^|/)shCore\.js$",
    "(^|/)shLegacy\.js$",
    # AngularJS
    "(^|/)angular([^.]*)\.js$",
    # D3.js
    "(^|\/)d3(\.v\d+)?([^.]*)\.js$",
    # React
    "(^|/)react(-[^.]*)?\.js$",
    # flow-typed
    "(^|/)flow-typed/.*\.js$",
    # Modernizr
    "(^|/)modernizr\-\d\.\d+(\.\d+)?\.js$",
    "(^|/)modernizr\.custom\.\d+\.js$",
    # Knockout
    "(^|/)knockout-(\d+\.){3}(debug\.)?js$",
    ## Python ##
    # Sphinx
    "(^|/)docs?/_?(build|themes?|templates?|static)/",
    # django
    "(^|/)admin_media/",
    "(^|/)local_settings\.py",
    # Flask
    "(^|/)instance/",
    # Fabric
    "(^|/)fabfile\.py$",
    # WAF
    "(^|/)waf$",
    # .osx
    "(^|/)\.osx$",
    ## Obj-C ##
    # Xcode
    ### these can be part of a directory name
    "\.xctemplate/",
    "\.imageset/",
    # Carthage
    "(^|/)Carthage/",
    # Sparkle
    "(^|/)Sparkle/",
    # Crashlytics
    "(^|/)Crashlytics\.framework/",
    # Fabric
    "(^|/)Fabric\.framework/",
    # BuddyBuild
    "(^|/)BuddyBuildSDK\.framework/",
    # Realm
    "(^|/)Realm\.framework",
    # RealmSwift
    "(^|/)RealmSwift\.framework",
    # git config files
    "(^|/)\.gitattributes$",
    "(^|/)\.gitignore$",
    "(^|/)\.gitmodules$",
    ## Groovy ##
    # Gradle
    "(^|/)gradlew$",
    "(^|/)gradlew\.bat$",
    "(^|/)gradle/wrapper/",
    ## Java ##
    # Maven
    "(^|/)mvnw$",
    "(^|/)mvnw\.cmd$",
    "(^|/)\.mvn/wrapper/",
    ## .NET ##
    # Visual Studio IntelliSense
    "-vsdoc\.js$",
    "\.intellisense\.js$",
    # jQuery validation plugin (MS bundles this with asp.net mvc)
    "(^|/)jquery([^.]*)\.validate(\.unobtrusive)?\.js$",
    "(^|/)jquery([^.]*)\.unobtrusive\-ajax\.js$",
    # Microsoft Ajax
    "(^|/)[Mm]icrosoft([Mm]vc)?([Aa]jax|[Vv]alidation)(\.debug)?\.js$",
    # NuGet
    "(^|/)[Pp]ackages\/.+\.\d+\/",
    # ExtJS
    "(^|/)extjs/.*?\.js$",
    "(^|/)extjs/.*?\.xml$",
    "(^|/)extjs/.*?\.txt$",
    "(^|/)extjs/.*?\.html$",
    "(^|/)extjs/.*?\.properties$",
    "(^|/)extjs/\.sencha/",
    "(^|/)extjs/docs/",
    "(^|/)extjs/builds/",
    "(^|/)extjs/cmd/",
    "(^|/)extjs/examples/",
    "(^|/)extjs/locale/",
    "(^|/)extjs/packages/",
    "(^|/)extjs/plugins/",
    "(^|/)extjs/resources/",
    "(^|/)extjs/src/",
    "(^|/)extjs/welcome/",
    # Html5shiv
    "(^|/)html5shiv\.js$",
    # Test fixtures
    "(^|/)[Tt]ests?/fixtures/",
    "(^|/)[Ss]pecs?/fixtures/",
    # PhoneGap/Cordova
    "(^|/)cordova([^.]*)\.js$",
    "(^|/)cordova\-\d\.\d(\.\d)?\.js$",
    # Foundation js
    "(^|/)foundation(\..*)?\.js$",
    # Vagrant
    "(^|/)Vagrantfile$",
    # R packages
    "(^|/)vignettes/",
    "(^|/)inst/extdata/",
    # Typesafe Activator
    "(^|/)activator$",
    # - (^|/)activator\.bat$
    # PuPHPet
    "(^|/)puphpet/",
    # Android Google APIs
    "(^|/)\.google_apis/",
    # Jenkins Pipeline
    "(^|/)Jenkinsfile$",
    # GitHub.com
    "(^|/)\.github/",
    # Environments
    "(^|/)venv/",
    "(^|/)env/",
    "(^|/)ENV/",
    "(^|/)env\.bak/",
    "(^|/)venv\.bak/",
    # Lock Files (NPM, Yarn, PNPM, etc.)
    "(^|/)package-lock\.json",
    "(^|/)yarn\.lock",
    "(^|/)pnpm-lock\.ya?ml",
    "(^|/)pnpm-workspace\.ya?ml",
    "(^|/)pnpm-workspace\.json",
    # Lock Files (Python)
    "(^|/)Pipfile\.lock",
    # Lock Files Generic
    "(^|/).*\.lock$",
]



LANGUAGE_EXTENSIONS = {
    ".py": "python",
    ".js": "javascript",
    ".mjs": "javascript",  # mjs file extension stands for "module JavaScript."
    ".go": "go",
    ".bash": "bash",
    ".c": "c",
    ".cc": "cpp",
    ".cs": "c_sharp",
    ".cl": "commonlisp",
    ".cpp": "cpp",
    ".css": "css",
    # ".dart": "dart", # py-tree-sitter-languages does not support Dart
    ".dockerfile": "dockerfile",
    ".dot": "dot",
    ".el": "elisp",
    ".ex": "elixir",
    ".elm": "elm",
    ".et": "embedded_template",
    ".erl": "erlang",
    ".gomod": "gomod",
    ".hack": "hack",
    ".hs": "haskell",
    ".hcl": "hcl",
    ".html": "html",
    ".java": "java",
    ".jsdoc": "jsdoc",
    ".json": "json",
    ".jl": "julia",
    ".kt": "kotlin",
    ".lua": "lua",
    ".mk": "make",
    # ".md": "markdown", # https://github.com/ikatyang/tree-sitter-markdown/issues/59
    ".m": "objc",
    ".ml": "ocaml",
    ".pl": "perl",
    ".php": "php",
    ".ql": "ql",
    ".r": "r",
    ".R": "r",
    ".regex": "regex",
    ".rst": "rst",
    ".rb": "ruby",
    ".rs": "rust",
    ".scala": "scala",
    ".sql": "sql",
    ".sqlite": "sqlite",
    ".tf": "hcl",
    ".toml": "toml",
    ".tsq": "tsq",
    ".tsx": "typescript",
    ".ts": "typescript",
    ".yaml": "yaml",
}



# Standard library
from collections import defaultdict

# Third party
import json_repair
from saplings.dtos import Message
from saplings.abstract import Tool
from litellm import acompletion, encode, decode

# Local
try:
    from suss.index import Index, File, Chunk
    from suss.constants import MAX_MERGE_DISTANCE, MAX_CONTEXT_TOKENS
except ImportError:
    from index import Index, File, Chunk
    from constants import MAX_MERGE_DISTANCE, MAX_CONTEXT_TOKENS


#########
# HELPERS
#########


PROMPT = """I want to find all the code in a file that's relevant to a query.

Your output should be a list of line ranges. Each line range should correspond to a block of code that's relevant to the query.

Line ranges should be inclusive (e.g. {{"start": 12, "end": 15}} includes lines 12, 13, 14, and 15).

--

Here is the code in the file ({file_path}):

<code>
{file_content}
</code>

And here is the query to find relevant code for:

<query>
{query}
</query>"""


def truncate_file_content(file: File) -> str:
    tokens = encode(model="", text=file.content)[:MAX_CONTEXT_TOKENS]
    file_str = decode(model="", tokens=tokens)
    return file_str


async def extract_chunks(file: File, query: str, model: str) -> list[Chunk]:
    file_path = file.rel_path
    file_content = truncate_file_content(file)
    messages = [
        {
            "role": "user",
            "content": PROMPT.format(
                file_path=file_path, file_content=file_content, query=query
            ),
        },
    ]
    response = await acompletion(
        model=model,
        messages=messages,
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "find_code_response",
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {
                        "line_ranges": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "start": {
                                        "type": "integer",
                                        "description": "Starting line number (inclusive).",
                                    },
                                    "end": {
                                        "type": "integer",
                                        "description": "Ending line number (inclusive).",
                                    },
                                },
                                "required": ["start", "end"],
                                "additionalProperties": False,
                            },
                            "description": "List of line ranges that contain relevant code.",
                        },
                    },
                    "required": ["line_ranges"],
                    "additionalProperties": False,
                },
            },
        },
        drop_params=True,
    )
    response = json_repair.loads(response.choices[0].message.content)
    line_ranges = [lr for lr in response["line_ranges"] if lr["start"] <= lr["end"]]
    line_ranges = [list(range(lr["start"], lr["end"] + 1)) for lr in line_ranges]
    chunks = [Chunk(line_nums, file) for line_nums in line_ranges]
    return chunks


def clamp_chunks(chunks: list[Chunk]) -> list[Chunk]:
    # Ensures that line numbers in chunks are within the file's range

    clamped_chunks = []
    for chunk in chunks:
        chunk.line_nums = [ln for ln in chunk.line_nums if ln <= chunk.file.last_lineno]
        if not chunk.line_nums:
            continue

        clamped_chunks.append(chunk)

    return clamped_chunks


def merge_chunks(chunks: list[Chunk]) -> list[Chunk]:
    merged_chunks = []
    if not chunks:
        return merged_chunks

    chunks.sort(key=lambda chunk: chunk.line_nums[0])
    curr_chunk = chunks[0]
    for next_chunk in chunks[1:]:
        curr_start, curr_end = curr_chunk.line_nums[0], curr_chunk.line_nums[-1]
        next_start, next_end = next_chunk.line_nums[0], next_chunk.line_nums[-1]

        is_overlapping = curr_end >= next_start
        is_within_distance = next_start - curr_end < MAX_MERGE_DISTANCE
        if is_overlapping or is_within_distance:
            curr_chunk.line_nums = list(range(curr_start, next_end + 1))
        else:
            merged_chunks.append(curr_chunk)
            curr_chunk = next_chunk

    merged_chunks.append(curr_chunk)
    merged_chunks = list(set(merged_chunks))

    return merged_chunks


######
# MAIN
######


class ReadFileTool(Tool):
    def __init__(
        self,
        index: Index,
        model: str,
        target_file: File,
        update_progress: callable,
        **kwargs,
    ):
        # Base attributes
        self.name = "read_file"
        self.description = "Read (search within) a file in the codebase. Returns the most relevant parts of the file."
        self.parameters = {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Concise, one-sentence description of your intent behind reading the file. E.g. 'Find the definition of handle_auth', 'Look for helper functions for the parser', etc.",
                },
                "file": {
                    "type": "string",
                    "enum": [],
                    "description": "The path to the file you want to read.",
                },
                "query": {
                    "type": "string",
                    "description": "A semantic search query. What are you looking for in the file?",
                },
            },
            "required": ["intent", "file", "query"],
            "additionalProperties": False,
        }
        self.is_terminal = False

        # Additional attributes
        self.index = index
        self.model = model
        self.target_file = target_file
        self.update_progress = update_progress

    def format_output(self, chunks: list[Chunk]) -> str:
        grouped_chunks = defaultdict(list)
        for chunk in chunks:
            grouped_chunks[chunk.file].append(chunk)

        formatted_chunks = []
        for file, chunks in grouped_chunks.items():
            for chunk in chunks:
                formatted_chunk = f"<{file.rel_path}>\n{chunk.to_string(dots=False)}\n</{file.rel_path}>"
                formatted_chunks.append(formatted_chunk)

        formatted_chunks = "\n\n".join(formatted_chunks)
        return formatted_chunks

    def update_definition(self, trajectory: list[Message] = [], **kwargs):
        files = set()
        for message in trajectory:
            if not message.raw_output:
                continue

            for item in message.raw_output:
                if isinstance(item, File):
                    files.add(item.rel_path)
                elif isinstance(item, Chunk):
                    files.add(item.file.rel_path)

        files = [file for file in files if file != self.target_file.rel_path]
        self.parameters["properties"]["file"]["enum"] = files

    def is_active(self, trajectory: list[Message] = [], **kwargs) -> bool:
        if not trajectory:
            return False

        for message in trajectory:
            if not message.raw_output:
                continue

            for item in message.raw_output:
                if isinstance(item, File):
                    self.update_definition(trajectory)
                    return True
                elif isinstance(item, Chunk):
                    self.update_definition(trajectory)
                    return True

        return False

    async def run(self, intent: str, file: str, query: str, **kwargs) -> list[Chunk]:
        self.update_progress(intent)
        file = self.index.get_file(file)
        chunks = await extract_chunks(file, query, self.model)
        chunks = clamp_chunks(chunks)
        chunks = merge_chunks(chunks)
        return chunks


# TODO: Allow for multiple files
# TODO: Alternative approach: chunk up each file, embed the chunks, then do vector search



# Standard library
import re
import os
import math

# Third party
import cohere
from tree_sitter_languages import get_parser

# Local
try:
    from suss.index import Chunk
except ImportError:
    from index import Chunk


#########
# HELPERS
#########


try:
    client = cohere.AsyncClient(os.environ["COHERE_API_KEY"])
except:
    client = None

MAX_CHUNKS = 10000


def tokenize_query(query: str) -> list[str]:
    tokens = re.findall(r"\w+", query)
    return tokens


# TODO: Make this less dumb
def tokenize_code(chunk: Chunk) -> list[str]:
    def extract_tokens(node, code_bytes: bytes) -> list[str]:
        tokens = []
        if node.child_count == 0:
            token = code_bytes[node.start_byte : node.end_byte].decode(
                "utf8", errors="ignore"
            )

            # Skip tokens that are only whitespace or punctuation
            if token.strip() and not re.fullmatch(r"\W+", token):
                tokens.append(token.lower())
        else:
            for child in node.children:
                tokens.extend(extract_tokens(child, code_bytes))

        return tokens

    parser = get_parser(chunk.file.language)
    code_bytes = bytes(chunk.to_string(False, False), "utf8")
    tree = parser.parse(code_bytes)
    return extract_tokens(tree.root_node, code_bytes)


def bm25_rerank(
    query: str, chunks: list[Chunk], k1: float = 1.5, b: float = 0.75
) -> list[Chunk]:
    if not chunks:
        return chunks

    tokenized_query = tokenize_query(query)
    tokenized_chunks = [tokenize_code(chunk) for chunk in chunks]
    N = len(tokenized_chunks)

    token_freq = {}
    for tokens in tokenized_chunks:
        for token in set(tokens):
            token_freq[token] = token_freq.get(token, 0) + 1

    avgdl = sum(len(tokens) for tokens in tokenized_chunks) / N

    # Compute BM25 scores
    scores = []
    for i, tokens in enumerate(tokenized_chunks):
        score = 0.0

        token_counts = {}
        for t in tokens:
            token_counts[t] = token_counts.get(t, 0) + 1

        dl = len(tokens)
        for t in tokenized_query:
            if t not in token_counts:
                continue

            n_t = token_freq.get(t, 0)
            idf = math.log((N - n_t + 0.5) / (n_t + 0.5) + 1)
            tf = token_counts[t]
            score += idf * (tf * (k1 + 1)) / (tf + k1 * (1 - b + b * (dl / avgdl)))

        scores.append((score, i))

    ranked_indices = sorted(scores, key=lambda x: x[0], reverse=True)
    ranked_chunks = [chunks[i] for _, i in ranked_indices]
    return ranked_chunks


async def neural_rerank(query: str, chunks: list[Chunk]) -> list[Chunk]:
    chunk_strs = [chunk.to_string(line_nums=False) for chunk in chunks[:MAX_CHUNKS]]
    response = await client.rerank(
        model="rerank-v3.5", query=query, documents=chunk_strs
    )
    for result in response.results:
        score = result.relevance_score
        index = result.index
        chunks[index].score = score

    ranked_chunks = sorted(chunks, key=lambda c: c.score, reverse=True)
    return ranked_chunks


######
# MAIN
######


async def rerank_chunks(query: str, chunks: list[Chunk]) -> list[Chunk]:
    if os.getenv("COHERE_API_KEY", None):
        return await neural_rerank(query, chunks)

    return bm25_rerank(query, chunks)



# Third party
from saplings.dtos import Message
from saplings import COTAgent, Model

# Local
try:
    from suss.tools import (
        GrepCodeTool,
        GlobFilesTool,
        ReadFileTool,
        FindBugsTool,
    )
    from suss.index import Index, File
except ImportError:
    from index import Index, File
    from tools import GrepCodeTool, GlobFilesTool, ReadFileTool, FindBugsTool


#########
# HELPERS
#########


SYSTEM_PROMPT = """<instructions>
Your job is to choose the best action. Call functions to find information about the \
codebase that will help you answer the user's query. Call functions.done when you \
have enough context to answer the user's query.
</instructions>

<rules>
- DO NOT call a function that you've used before with the same arguments.
- DO NOT assume the structure of the codebase or the existence of other files \
or folders.
- Your queries to functions.search_code and functions.search_files should be \
significantly different than previous queries.
- If the output of a function is empty, try calling the function again with \
DIFFERENT arguments OR try calling a different function.
</rules>"""


def build_prompt(file: File) -> str:
    prompt = f"<path>{file.path}</path>\n"
    prompt += f"<code>\n{file.content}\n</code>\n\n"
    prompt += "--\n\nAbove is a file from the codebase. Analyze it for bugs. Consider bugs that are isolated to the file, and bugs it may cause in other files (i.e. afferent or efferent dependencies)."
    return prompt


def was_tool_called(messages: list[Message], tool_name: str) -> bool:
    for message in messages:
        if message.role != "assistant":
            continue

        if not message.tool_calls:
            continue

        for tool_call in message.tool_calls:
            if tool_call.name == tool_name:
                return True

    return False


######
# MAIN
######


class Agent:
    def __init__(self, index: Index, model: str, max_iters: int):
        self.index = index
        self.model = model
        self.max_iters = max_iters

    async def run(self, file: File, update_progress: callable):
        # TODO: Add pseudo-semantic search tool
        # TODO: Add call graph tools (ref/def search)
        tools = [
            GrepCodeTool(self.index, file, update_progress),
            GlobFilesTool(self.index, file, update_progress),
            ReadFileTool(self.index, self.model, file, update_progress),
            FindBugsTool(self.model, file, update_progress),
        ]
        model = Model(self.model)
        agent = COTAgent(
            tools,
            model,
            SYSTEM_PROMPT,
            tool_choice="required",
            max_depth=self.max_iters,
            verbose=False,
        )
        prompt = build_prompt(file)
        messages = await agent.run_async(prompt)

        output = messages[-1].raw_output
        if not was_tool_called(messages, "done"):
            tool_call = await agent.call_tool("done", messages)
            tool_result = await agent.run_tool(tool_call, messages)
            output = tool_result.raw_output

        return output



import sys

from setuptools import setup, find_packages

if sys.version_info[:3] < (3, 0, 0):
    print("Requires Python 3 to run.")
    sys.exit(1)

setup(
    name="suss",
    version="0.1.0",
    description="AI-powered bug finder that knows your codebase",
    author="shobrook",
    author_email="shobrookj@gmail.com",
    url="https://github.com/shobrook/suss",
    packages=find_packages(),
    python_requires=">=3.12",
    install_requires=[
        "cohere",
        "json-repair",
        "litellm",
        "rich",
        "saplings",
        "sortedcollections",
        "tree-sitter",
        "tree-sitter-languages",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
    ],
    # long_description=open("README.md").read(),
    keywords="code-review bug-finder ai debugger agent llm codebase code-assistant",
    entry_points={
        "console_scripts": [
            "suss=suss.suss:main",
        ],
    },
)
